
class Signal; class Zone;

/////////////////////////////////////////////////////////////////////////////

class Aiguille {
  
protected:
  boolean etat; // directe (true) ou deviee (false)
  int no;       // numéro de l'aiguille
  Zone* zone;   // la Zone ou est l'aiguille
  
public: 
  Aiguille(int n,Zone* z) { no=n; zone=z; } // constructeur
  
  boolean directe() { return etat; }  // méthode d'acces
  boolean deviee()  { return !etat; } // méthode d'acces
  
  boolean directer(); // manoeuvre de l'aiguille si necessaire
  boolean devier();   // manoeuvre de l'aiguille si necessaire
  
  boolean manoeuvrable(); //  { return zone->libre(); }
  boolean manoeuvrer(boolean e); // interface le bus
  
  void init1(boolean e) { etat=e; }
  void init2(boolean e) { if (e) manoeuvrer(e); else manoeuvrer(!e);  etat=e; }
};

//  //  // //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //

  boolean Aiguille::directer() { if (directe()) return true; etat=true;  return manoeuvrer(true); }
  boolean Aiguille::devier()   { if (deviee())  return true; etat=false; return manoeuvrer(false); }

  boolean Aiguille::manoeuvrer(boolean e) { // commande de l'aiguille avec le numero  
   return true; // return false si cela c'est mal passe
 }

/////////////////////////////////////////////////////////////////////////////

class Zone { //  protections acces ???
public:
   boolean etat; // libre ou occupé
   Signal* signalPair; Signal* signalImpair; // les signaux éventuels de la zone

   Zone(Signal* sp,Signal* si) { signalPair=sp; signalImpair=si; } // constructeur
   
   boolean occupee() { return etat; }  // méthode d'acces
   boolean libre()   { return !etat; } // méthode d'acces

   virtual Zone* suivantePaire()   { return NULL; } // la zone suivante paire (éventuellement vide)
   virtual Zone* suivanteImpaire() { return NULL; } // la zone suivante impaire (éventuellement vide)

   virtual void actions() {} // les actions spécifiques à faire en cas d'occupation
   virtual void desactions() {} // les actions spécifiques à faire en cas de libération

   void occuper() { // appelée par la rétrosignalisation
      // fait tout ce qu'il y a à faire en cas d'occupation (actions communes à toutes les zones)
      actions(); // fait les actions spécifiques à une zone
   }
   void liberer() { // appelée par la rétrosignalisation
      // fait tout ce qu'il y a à faire en cas de libération (actions communes à toutes les zones)
      desactions(); // fait les actions spécifiques à une zone
   }

   Zone* selonAiguille(Aiguille* a,Zone* z1,Zone* z2) { return a->directe()?z1:z2; } // méthode utilitaire
};

//  //  // //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //


/////////////////////////////////////////////////////////////////////////////

typedef unsigned int typeFeux; // 16 cas possibles   !!! variantes

enum Feu {Vl=1,   A=2,   S=4,C=8,R=16,   RR=32,   M=64,Cv=128,                  // puissances de 2
          Vlc=256,Ac=512,Sc=1024,Rc=2048,RRc=4096,Mc=8192,    D=16384,X=32768}; // puissances de 2
    
class Signal {
protected:
  typeFeux feu; 
  int no; // numero du signal
public:
  Signal(typeFeux f,int n) { feu=f; no=n; } // constructeur
  
  typeFeux feux() { return feu; } // methode d'acces    !!! probleme avec le nom
  boolean ferme() { return feu&(S|C|Cv); }
  boolean ouvert() { return !ferme(); }
  
  virtual boolean ouvrir()=0; // ouverture manuelle
  virtual boolean fermer()=0; // fermeture manuelle
  virtual boolean aubiner() { return fermer(); } // fermeture par retrosignalisation

  boolean manoeuvrer() {/* commande du signal en utilisant feu et no */ return false; /* true si OK */ }

  virtual boolean ralentissement30() { return false; }
  virtual boolean ralentissement60() { return false; }

  virtual boolean cantonOccupe() { return false; } // occupation du canton pour BAL

  boolean v30() { return feu&RR; } // vitesse 30
  boolean r30() { return feu&A|R; } // ralentissement a 30

  boolean oeuilleton() { if (feu&(Vl+A+S+R+RR)!=0) return true; else return false; } // methode utilitaire
  boolean clignotant() { if (feu&(Vlc+Ac+Sc+Rc+RRc+Mc)!=0) return true; else return false; } // methode utilitaire

public:
  virtual void maj() {} // mise a jour du feu precedent
  
  virtual Signal* suivant()   { return NULL; }
  virtual Signal* precedent() { return NULL; }
  
// methodes utilitaires
  Signal* selonAiguille(Aiguille* a,Signal* s1,Signal* s2) { return a->directe()?s1:s2; } 

  boolean occupee(Zone* z) { return z->occupee(); }
  boolean occupees(Zone* z1,Zone* z2) { return z1->occupee() || z2->occupee(); }
  boolean occupees(Zone* z1,Zone* z2,Zone* z3) { return z1->occupee() || z2->occupee() || z3->occupee(); }
};

// signaux sans BAL

class CarreViolet:public Signal {
  CarreViolet(int n):Signal(Cv,n) {} // constructeur
  
  boolean ouvrir() { if (ouvert()) return false; feu=M;  return manoeuvrer(); }
  boolean fermer() { if (ferme())  return false; feu=Cv; return manoeuvrer(); }
};

class Avertissement:public Signal { // le suivant est forcement un carre
public:
  Avertissement(int n):Signal(A,n) {} // constructeur    ouvert ???

  boolean ouvrir() {        /* manoeuvrer() */ return true; } // feu=Vl ou maj() ou rien
  boolean fermer() { if (ferme()) return false; feu=A; return manoeuvrer(); }

  virtual void maj() { feu=suivant()->ferme()?A:Vl; } // appele par le carre suivant
};

class AvertissementRalentissement:public Avertissement { // le suivant est forcement un carre
public:
  AvertissementRalentissement(int n):Avertissement(n) {} // constructeur 
  
  virtual void maj() { // appele par le carre suivant
    if (suivant()->ferme())            feu=A; else
    if (suivant()->ralentissement60()) feu=Rc; else    // revoir les maj()) : comande signal !!!
    if (suivant()->ralentissement30()) feu=R;
    else                               feu=Vl;
  }  
};

class Carre:public Signal {
public:
  Carre(int n):Signal(C,n) {} // constructeur
  boolean ouvrir() { if (ouvert()) return false; feu=feux(); precedent()->maj(); return manoeuvrer(); }
  boolean fermer() { if (ferme())  return false; feu=C;      precedent()->maj(); return manoeuvrer(); }

  virtual typeFeux feux() { // feux a l'ouverture
    if (suivant()->ferme()) return A; else return Vl; // suivant null ???
  }

  virtual void maj() { typeFeux f; // appele par le signal suivant
    if (ferme()) return; // pas de propagation si pas de changement
    f=feux(); if (f!=feu) { feu=f; precedent()->maj(); } 
  }
};

class CarreRappelRalentissement:public Carre {
public:
  CarreRappelRalentissement(int n):Carre(n) {} // constructeur
  virtual typeFeux feux() { // feux a l'ouverture
    if (ralentissement60()) if (suivant()->ferme()) return A+RRc; else return RRc; else
    if (ralentissement30()) if (suivant()->ferme()) return A+RR;  else return RR; 
    else                    if (suivant()->ferme()) return A;     else return Vl;
  }
};

class CarreRalentissement:public Carre {
public:
  CarreRalentissement(int n):Carre(n) {} // constructeur
  virtual typeFeux feux() { // feux a l'ouverture
    if (suivant()->ralentissement60()) if (suivant()->ferme()) return A+RRc; else return RRc; else
    if (suivant()->ralentissement30()) if (suivant()->ferme()) return A+RR;  else return RR; 
    else                               if (suivant()->ferme()) return A;     else return Vl;
  }
};

class CarreRalentissementRappelRalentissement:public Carre { // tableau
public:
  CarreRalentissementRappelRalentissement(int n):Carre(n) {} // constructeur
  virtual typeFeux feux() { // feux a l'ouverture
    if (suivant()->ferme()) 
      if (ralentissement60()) return A+RRc; else  
      if (ralentissement30()) return A+RR;   
      else                    return A; 
    else // suivant ouvert
      if (suivant()->ralentissement60()) return RRc; else  
      if (suivant()->ralentissement30()) return RR;   
      else                                                             return Vl; 
  }
};

class Semaphore:public Signal { // avertissement ???
public:
  Semaphore(int n):Signal(S,n) {} // constructeur 
  
  boolean ouvrir() { if (ouvert()) return false; feu=Vl; precedent()->maj(); return manoeuvrer(); }
  boolean fermer() { if (ferme())  return false; feu=S;  precedent()->maj(); return manoeuvrer(); }
};

class Disque:public Signal { // avertissement ???  feu ouvert ???
public:
  Disque(int n):Signal(D,n) {} // constructeur 
  
  boolean ouvrir() { if (ouvert()) return false; feu=Vl; precedent()->maj(); return manoeuvrer(); }
  boolean fermer() { if (ferme())  return false; feu=S;  precedent()->maj(); return manoeuvrer(); }
};

// les signaux avec BAL

class SemaphoreBAL:public Signal { //  pas fermer mais aubiner
public:  
  SemaphoreBAL(boolean a,int n):Signal(a?A:Vl,n) {} // constructeur (si a=true : avertissement de carre)
  SemaphoreBAL(int n):SemaphoreBAL(false,n) {} // autre constructeur
  
  boolean ouvrir() { feu=feux(); precedent()->maj(); return manoeuvrer(); }
  boolean fermer() { return false; } // pas de commande manuelle
  boolean aubiner() { if (ferme()) return false; feu=S; precedent()->maj(); return manoeuvrer(); }

  typeFeux feux() { // feux a l'ouverture
    if (cantonOccupe()) return S; else // canton occupe ( precaution )
    if (suivant()->ferme()) return A; else return Vl;
  }

  void maj() { typeFeux f; // appele par le signal suivant            commande signal !!!
    if (ferme()) return; // pas de propagation si pas de changement
    f=feux(); if (f!=feu) { feu=f; precedent()->maj(); } 
  }
};

class SemaphoreBALRalentissement:public SemaphoreBAL {
public:
  SemaphoreBALRalentissement(int n):SemaphoreBAL(true,n) {} // constructeur (le signal suivant est forcement un carre)
  
  virtual typeFeux feux() { // feux a l'ouverture
    if (suivant()->ferme()) feu=A;   else 
    if (ralentissement60()) feu=RRc; else
    if (ralentissement30()) feu+=RR; 
    else                    feu=Vl;  
  }
};

class CarreBAL:public Signal {
  boolean ouverturePermanente=false;
public:
  CarreBAL(boolean p,int n):Signal(C,n) { ouverturePermanente=p; } // constructeur
  CarreBAL(int n):Signal(C,n) { ouverturePermanente=false; } // autre constructeur
  
  boolean ouvrir() { if (ouvert()) return false; feu=feux(); precedent()->maj(); return manoeuvrer(); }
  boolean fermer() { if (ferme())  return false; feu=C; precedent()->maj(); return manoeuvrer(); }
  boolean aubiner() { return false; }

  void permanent(boolean p) { ouverturePermanente=p; } // noms !!!
  
  virtual typeFeux feux() { // feux a l'ouverture
    if (suivant()->ferme()) return A; else return Vl; // suivant null ???
  }

  virtual void maj() { typeFeux f; // appele par le signal suivant
    if (ferme()) return; // pas de propagation si pas de changement
    f=feux(); if (f!=feu) { feu=f; precedent()->maj(); } 
  }
};

class CarreBALRappelRalentissement:public CarreBAL {
public:
  CarreBALRappelRalentissement(int n):CarreBAL(n) {} // constructeur
  virtual typeFeux feux() { // feux a l'ouverture
    if (ralentissement60()) if (suivant()->ferme()) return A+RRc; else return RRc; else
    if (ralentissement30()) if (suivant()->ferme()) return A+RR;  else return RR; 
    else                    if (suivant()->ferme()) return A;     else return Vl;
  }
};

class CarreBALRalentissement:public CarreBAL {
public:
  CarreBALRalentissement(int n):CarreBAL(n) {} // constructeur
  virtual typeFeux feux() { // feux a l'ouverture     tester ferme avant !!!
    if (suivant()->ralentissement60()) if (suivant()->ferme()) return A+Rc; else return Rc; else
    if (suivant()->ralentissement30()) if (suivant()->ferme()) return A+R;  else return R; 
    else                               if (suivant()->ferme()) return A;    else return Vl;
  }
};

class CarreBALRalentissementRappelRalentissement:public CarreBAL {}; // a faire

// carres ouverture permanente

/////////////////////////////////////////////////////////////////////////////
// les zones reelles

class Z8 : public Zone { // héritage de Zone
public:
  Z8(Signal* sp,Signal* si):Zone(sp,si) {}
  virtual Zone* suivantePaire();
  virtual Zone* suivanteImpaire();
};
class Z9 : public Zone { // héritage de Zone
public:
  Z9(Signal* sp,Signal* si):Zone(sp,si) {}
  virtual Zone* suivantePaire();
  virtual Zone* suivanteImpaire();
};

/////////////////////////////////////////////////////////////////////////////
// les signaux reels

class C2:public Carre {
public:
  C2(int n):Carre(n) {} // constructeur
  virtual Signal* suivant();
  virtual Signal* precedent();
};

class S2:public SemaphoreBAL {
public:
  S2(boolean a,int n):SemaphoreBAL(a,n) {} // constructeur
  
  virtual Signal* suivant();
  virtual Signal* precedent();
  virtual boolean occupe();
};

/////////////////////////////////////////////////////////////////////////////
// declarations

Signal* c2=new C2(3); Signal* s2=new S2(true,5); // signaux

Zone* z8=new Z8(s2,c2); Zone* z9=new Z9(s2,NULL); // zones

Zone* tableZones[] {z8,z9}; // la table des zones pour la retrosignalisation

Aiguille* a1=new Aiguille(1,z8); Aiguille* a2=new Aiguille(2,z9); // aiguilles

/////////////////////////////////////////////////////////////////////////////
// les methodes des zones

Zone* Z8::suivantePaire() { return selonAiguille(a1,z8,z9); }
Zone* Z8::suivanteImpaire() { return z9; }

Zone* Z9::suivantePaire() { return z8; }
Zone* Z9::suivanteImpaire() { return NULL; }

/////////////////////////////////////////////////////////////////////////////
// les methodes des signaux

  Signal* C2::suivant() { return selonAiguille(a2,c2,s2); } // selon a2
  Signal* C2::precedent() { return s2; }

  Signal* S2::suivant() { return s2; } 
  Signal* S2::precedent() { return c2; }
  boolean S2::occupe() { return occupees(z8,z9); } // occupation du canton (2 zones)

/////////////////////////////////////////////////////////////////////////////

class Itineraire { // classe de base des itineraires
  byte etat; // l'etat de l'itineraire (libre, en attente, forme, ... )

public:
  virtual boolean formable();
  virtual boolean former();
  virtual boolean deformable();
  virtual boolean deformer();

  boolean libre() { return etat==0; }
  boolean formation()   { if (formable()) return former(); else return false; }
  boolean destruction() { if (deformable()) return deformer(); else return false; }
   
// methodes utilitaires
  boolean libre(Zone* a) { return a->libre(); }
  boolean libres(Zone* a,Zone* b) { return a->libre() && b->libre(); }
  boolean libres(Zone* a,Zone* b,Zone* c) { return a->libre() && b->libre()&& c->libre(); }
  boolean libres(Zone* a,Zone* b,Zone* c,Zone* d) { return a->libre() && b->libre() && c->libre() && d->libre(); }
  
  boolean libre(Itineraire* a) { return a->libre(); }
  boolean libres(Itineraire* a,Itineraire* b) { return a->libre() && b->libre(); }
  boolean libres(Itineraire* a,Itineraire* b,Itineraire* c) { return a->libre() && b->libre()&& c->libre(); }
  boolean libres(Itineraire* a,Itineraire* b,Itineraire* c,Itineraire* d) { return a->libre() && b->libre() && c->libre() && d->libre(); }
  // ...
  };

/////////////////////////////////////////////////////////////////////////////
// les itineraires reels

class TC:public Itineraire { // un itineraire particulier de T vers C
  virtual boolean formable();
  virtual boolean former() { return a1->devier() && a2->directer() && c2->ouvrir(); }
  virtual boolean deformable() { return libres(z8,z9); }
  virtual boolean deformer() { return c2->fermer(); }
};

class CT:public Itineraire { // un itineraire particulier de C vers T
  virtual boolean formable();
  virtual boolean former() { return a1->devier() && a2->directer() && c2->ouvrir(); }
  virtual boolean deformable() { return libres(z8,z9); }
  virtual boolean deformer() { return c2->fermer(); }
};

/////////////////////////////////////////////////////////////////////////////
// declarations

Itineraire* iTC=new TC(); Itineraire* iCT=new CT(); // itineraires

/////////////////////////////////////////////////////////////////////////////
// methodes des itineraires

boolean TC::formable() { return 
    libres(iTC,iCT) && // enclanchements entre itineraires
    libres(z8,z9); // enclanchements sur les zones
}
 
boolean CT::formable() { return 
    libres(iTC,iCT) && // enclanchements entre itineraires
    libres(z8,z9); // enclanchements sur les zones
}

/////////////////////////////////////////////////////////////////////////////

void setup() {
  // put your setup code here, to run once:

}

void loop() {
  // put your main code here, to run repeatedly:

}
