
// classe de base des signaux

/////////////////////////////////////////////////////////////////////////////

typedef unsigned int typeFeux; // 16 cas possibles ( 32 avec long )

// les feux possibles
enum {
  E=0,        // eteint
  Vl = 1<<0,   // 1 ou 0b0000000000000001
  A  = 1<<1,   // 2 ou 0b0000000000000010
  S  = 1<<2,   // 4 ou 0b0000000000000100
  C  = 1<<3,   // 8 ou 0b0000000000001000
  R  = 1<<4,   // 16 ou 0b0000000000010000
  RR = 1<<5,   // 32 ou 0b0000000000100000
  M  = 1<<6,   // 64 ou 0b0000000001000000
  Cv = 1<<7,   // 128 ou 0b0000000010000000
  Vlc= 1<<8,   // 256 ou 0b0000000100000000
  Ac = 1<<9,   // 512 ou 0b0000001000000000
  Sc = 1<<10,  // 1024 ou 0b0000010000000000
  Rc = 1<<11,  // 2048 ou 0b0000100000000000
  RRc= 1<<12,  // 4096 ou 0b0001000000000000
  Mc = 1<<13,  // 8192 ou 0b0010000000000000
  D  = 1<<14,  // 16384 ou 0b0100000000000000
  X  = 1<<15   // 32768 ou 0b1000000000000000
}; 

enum {  // les cibles   OPTION
  RIEN      = 0,
  CIBLE     = bit(0),   // 1 ou 0b0000000000000001
  CIBLE_BIS = bit(1),   // 2 ou 0b0000000000000010
  CIBLE_TER = bit(2),   // 4 ou 0b0000000000000100
}; 

enum {  // des types de signaux   OPTION
  MECANIQUE               =1<<0,   // 1 ou 0b0000000000000001
  LUMINEUX                =1<<1,   // 2 ou 0b0000000000000010
  AVERTISSEMENT           =1<<2,   // 4 ou 0b0000000000000100
  SEMAPHORE               =1<<3,   // 8 ou 0b0000000000001000
  CARRE                   =1<<4,   // 16 ou 0b0000000000010000
  VIOLET                  =1<<5,   // 32 ou 0b0000000000100000 
  RALENTISSEMENT          =1<<6,   // 64 ou 0b0000000001000000
  RALENTISSEMENT30        =1<<7,   // 128 ou 0b0000000010000000
  RAPPEL_RALENTISSEMENT   =1<<8,   // 256 ou 0b0000000100000000
  RAPPEL_RALENTISSEMENT30 =1<<9,   // 512 ou 0b0000001000000000
  RALENTISSEMENT60        =1<<10,  // 1024 ou 0b0000010000000000
  RAPPEL_RALENTISSEMENT60 =1<<11,  // 2048 ou 0b0000100000000000
  DISQUE                  =1<<12,  // 4096 ou 0b0001000000000000
  BAS                     =1<<13,  // 8192 ou 0b001000000000000
  BAL                     =1<<14   // 16384 ou 0b0100000000000000
};

class Signal {
protected:
  typeFeux feux,feuxInit; 
public:
  int no; // le numero du signal pour commander l'electronique ( feux, moteurs, ... )
  char* nom; // le nom du signal   OPTION
  int type; // le type du signal   OPTION
protected:   
  Signal(typeFeux f,int n) { feux=feuxInit=f; no=n; } // constructeur mini
  Signal(int t,typeFeux f,int n) { feux=feuxInit=f; no=n; } // constructeur 
  Signal(char* nm,int t,typeFeux f,int n) { nom=nm; type=t; feux=feuxInit=f; no=n; } // constructeur maxi
public:  
  typeFeux getFeux() { return feux; } // methode d'acces    
  
  boolean ferme() { return feux&(S|C|Cv); } // methode de test
  boolean ouvert() { return !ferme(); }     // methode de test
  
  virtual void ouvrir() {} // ouverture manuelle 
  virtual void fermer() {} // fermeture manuelle 
  virtual void aubiner() { fermer(); } // fermeture automatique par retrosignalisation
  
  void manoeuvrer(); // manoeuvre effective du signal (feux et/ou cibles)

  virtual boolean ralentissement30() { return false; } // true si ralentissement a 30
  virtual boolean ralentissement60() { return false; } // true si ralentissement a 60

  virtual void majFeux() {} // mise a jour du signal, appelle par le signal suivant
  
  virtual Signal* suivant()   { return NULL; } // le signal suivant  
  virtual Signal* precedent() { return NULL; } // le signal precedent  
  
 void init() { feux=feuxInit; } // (re)initialisation

 void erreur(int n); // erreur sur les signaux (tranche 300)
  
// methodes utilitaires

  // boolean clignotant() { return (feux&(Rc+RRc)); } // methode utilitaire   OPTION
  
  virtual int cibles() { return ouvert()?RIEN:CIBLE; } // les cibles a presenter   OPTION
  
  Signal* selonAiguille(Aiguille* a,Signal* s1,Signal* s2) { return a->directe()?s1:s2; } 

  boolean occupee(Zone* z); //   OPTION
  boolean occupees(Zone* z1,Zone* z2); //   OPTION
  boolean occupees(Zone* z1,Zone* z2,Zone* z3); //   OPTION

protected:
  void manoeuvreEtMajFeux(); // manoeuvre le signal puis fait la mise a jour du precednet
};

//  //  // //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //

void Signal::manoeuvrer() { // appele quand un signal change (feux et/ou cible)
  // COMPLETER manoeuvrer le signal (s'il existe ) ( utiliser l'entier feu et l'entier no ( eventuellement l'entier type ))
  // gerer la zone d'arret si elle existe
}

void Signal::manoeuvreEtMajFeux() { Signal* s;
  manoeuvrer(); 
  s=precedent(); if (s==NULL) { erreur(302); return; } s->majFeux(); 
}

void Signal::erreur(int n) { /* COMPLETER */ }

inline boolean Signal::occupee(Zone* z) { return z->occupee(); } 
       boolean Signal::occupees(Zone* z1,Zone* z2) { return z1->occupee() || z2->occupee(); } 
       boolean Signal::occupees(Zone* z1,Zone* z2,Zone* z3) { return z1->occupee() || z2->occupee() || z3->occupee(); } 

/////////////////////////////////////////////////////////////////////////////

